/*
 * Created on Aug 25, 2009
 * Created by Paul Gardner
 * 
 * Copyright 2009 Vuze, Inc.  All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2 of the License only.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */


package com.aelitis.azureus.plugins.offlinedownloader;

import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.net.InetAddress;
import java.net.NetworkInterface;
import java.net.URL;
import java.util.*;

import org.gudy.azureus2.core3.config.COConfigurationManager;
import org.gudy.azureus2.core3.disk.DiskManager;
import org.gudy.azureus2.core3.disk.DiskManagerPiece;
import org.gudy.azureus2.core3.peer.PEPeerManager;
import org.gudy.azureus2.core3.torrent.TOTorrent;
import org.gudy.azureus2.core3.torrent.TOTorrentAnnounceURLSet;
import org.gudy.azureus2.core3.torrent.TOTorrentFactory;
import org.gudy.azureus2.core3.util.AERunnable;
import org.gudy.azureus2.core3.util.AESemaphore;
import org.gudy.azureus2.core3.util.AsyncDispatcher;
import org.gudy.azureus2.core3.util.BDecoder;
import org.gudy.azureus2.core3.util.ByteFormatter;
import org.gudy.azureus2.core3.util.Constants;
import org.gudy.azureus2.core3.util.Debug;
import org.gudy.azureus2.core3.util.SystemTime;
import org.gudy.azureus2.core3.util.TorrentUtils;
import org.gudy.azureus2.plugins.Plugin;
import org.gudy.azureus2.plugins.PluginConfig;
import org.gudy.azureus2.plugins.PluginInterface;
import org.gudy.azureus2.plugins.PluginListener;
import org.gudy.azureus2.plugins.download.Download;
import org.gudy.azureus2.plugins.download.DownloadManager;
import org.gudy.azureus2.plugins.download.DownloadManagerListener;
import org.gudy.azureus2.plugins.download.DownloadPeerListener;
import org.gudy.azureus2.plugins.logging.LoggerChannel;
import org.gudy.azureus2.plugins.logging.LoggerChannelListener;
import org.gudy.azureus2.plugins.peers.PeerManager;
import org.gudy.azureus2.plugins.torrent.TorrentAttribute;
import org.gudy.azureus2.plugins.torrent.TorrentManager;
import org.gudy.azureus2.plugins.tracker.Tracker;
import org.gudy.azureus2.plugins.tracker.web.TrackerWebContext;
import org.gudy.azureus2.plugins.tracker.web.TrackerWebPageGenerator;
import org.gudy.azureus2.plugins.tracker.web.TrackerWebPageRequest;
import org.gudy.azureus2.plugins.tracker.web.TrackerWebPageResponse;
import org.gudy.azureus2.plugins.ui.UIManager;
import org.gudy.azureus2.plugins.ui.config.BooleanParameter;
import org.gudy.azureus2.plugins.ui.config.Parameter;
import org.gudy.azureus2.plugins.ui.config.ParameterListener;
import org.gudy.azureus2.plugins.ui.config.StringParameter;
import org.gudy.azureus2.plugins.ui.model.BasicPluginConfigModel;
import org.gudy.azureus2.plugins.ui.model.BasicPluginViewModel;
import org.gudy.azureus2.plugins.utils.DelayedTask;
import org.gudy.azureus2.plugins.utils.LocaleUtilities;
import org.gudy.azureus2.plugins.utils.UTTimer;
import org.gudy.azureus2.plugins.utils.UTTimerEvent;
import org.gudy.azureus2.plugins.utils.UTTimerEventPerformer;
import org.gudy.azureus2.plugins.utils.Utilities;
import org.gudy.azureus2.plugins.utils.resourcedownloader.ResourceDownloaderFactory;
import org.gudy.azureus2.plugins.utils.xml.simpleparser.SimpleXMLParserDocument;
import org.gudy.azureus2.plugins.utils.xml.simpleparser.SimpleXMLParserDocumentException;
import org.gudy.azureus2.plugins.utils.xml.simpleparser.SimpleXMLParserDocumentNode;
import org.gudy.azureus2.pluginsimpl.local.PluginCoreUtils;
import org.gudy.azureus2.pluginsimpl.local.peers.PeerManagerImpl;

import com.aelitis.azureus.core.networkmanager.admin.NetworkAdmin;
import com.aelitis.azureus.core.peermanager.piecepicker.PiecePicker;
import com.aelitis.azureus.core.peermanager.piecepicker.PiecePriorityProvider;
import com.aelitis.azureus.core.util.UUIDGenerator;

import com.aelitis.net.upnp.*;


public class 
OfflineDownloaderPlugin 
	implements Plugin
{
	private static final String CATEGORY_SERVER_ONLY = "server-only";
	
	private static final int	ACTIVE_CHECK_PERIOD	= 30*1000;
	private static final int 	ACTIVE_TIMEOUT		= 2*60*1000;
	
	private static final int	UPNP_ALIVE_PERIOD	= 60*1000;
	
	private static URL	DUMMY_ANNOUNCE_URL;
	
	static{
		try{
			DUMMY_ANNOUNCE_URL = new URL( "http://inactive/" );
			
		}catch( Throwable e ){
		
			Debug.out( e );
			
			DUMMY_ANNOUNCE_URL = null;
		}
	}
	private PluginInterface		plugin_interface;
	private LoggerChannel		logger;
	
	private TorrentAttribute	ta_category;
	private TorrentAttribute	ta_announce_cache;
	private TorrentAttribute	ta_piece_details;
	
	private String				service_name	= "Vuze Offline Downloader";
	private String				UUID_rootdevice;
	
	private String[]			upnp_entities;
	
	private boolean					alive = true;
	
	private UPnPSSDP 				ssdp;
	private UPnPSSDPListener 		ssdp_listener;
	
	private boolean		enable_upnp		= true;
	private Object		startup_lock	= new Object();
	private boolean		starting		= false;
	private AESemaphore	startup_sem 	= new AESemaphore( "OfflineDownloader" );

	private AsyncDispatcher	async_dispatcher = new AsyncDispatcher();
	
	private Map<String,clientState>	active_clients = new HashMap<String,clientState>();
	
	private Map<Download,long[]>	download_priorities_map	= new HashMap<Download, long[]>();
	private Map<Download,long[]>	xfer_priorities_map		= new HashMap<Download, long[]>();
	
	
	private Set<Download>			explicit_starts = new HashSet<Download>();
	
	private boolean					slave_mode;
	
	public void
	initialize(
		PluginInterface		_plugin_interface )
	{
		try{	
			COConfigurationManager.setParameter( "Confirm Data Delete", false );
			COConfigurationManager.setParameter( "Move Deleted Data To Recycle Bin", false );
			
			plugin_interface		= _plugin_interface;
						
			logger				= plugin_interface.getLogger().getTimeStampedChannel( "OfflineDownloader" ); 
	
			TorrentManager torrent_manager = plugin_interface.getTorrentManager();
			
			ta_category	= torrent_manager.getAttribute( TorrentAttribute.TA_CATEGORY );
			
			ta_announce_cache = torrent_manager.getPluginAttribute( "announce_cache" );
			
			ta_piece_details = torrent_manager.getPluginAttribute( "piece_details" );
			
			LocaleUtilities loc_utils = plugin_interface.getUtilities().getLocaleUtilities();
	
			loc_utils.integrateLocalisedMessageBundle( "com.aelitis.azureus.plugins.offlinedownloader.internat.Messages" );
												
			PluginConfig	config = plugin_interface.getPluginconfig();
			
			UUID_rootdevice = config.getPluginStringParameter( "uuid", "" );
			
			if ( UUID_rootdevice.length() == 0 ){
				
				UUID_rootdevice = "uuid:" + UUIDGenerator.generateUUIDString();
			
				config.setPluginParameter( "uuid", UUID_rootdevice );
			}
		
			upnp_entities = new String[]{
					"upnp:rootdevice",
					"urn:schemas-upnp-org:device:VuzeOfflineDownloader:1",
					UUID_rootdevice
				};
			
			final UIManager	ui_manager = plugin_interface.getUIManager();
			
			final BasicPluginViewModel view_model = ui_manager.createBasicPluginViewModel( "offlinedownloader.name" );
			
			view_model.getActivity().setVisible( false );
			view_model.getProgress().setVisible( false );
			
			logger.addListener(
					new LoggerChannelListener()
					{
						public void
						messageLogged(
							int		type,
							String	content )
						{
							view_model.getLogArea().appendText( content + "\n" );
						}
						
						public void
						messageLogged(
							String		str,
							Throwable	error )
						{
							if ( str.length() > 0 ){
								view_model.getLogArea().appendText( str + "\n" );
							}
							
							StringWriter sw = new StringWriter();
							
							PrintWriter	pw = new PrintWriter( sw );
							
							error.printStackTrace( pw );
							
							pw.flush();
							
							view_model.getLogArea().appendText( sw.toString() + "\n" );
						}
					});		
		
			logger.setDiagnostic();
			
			view_model.setConfigSectionID( "offlinedownloader.name" );
			
			logger.log( "RootDevice: " + UUID_rootdevice );
			
			BasicPluginConfigModel config_model = 
				ui_manager.createBasicPluginConfigModel( "offlinedownloader.name" );
			
			final StringParameter sn_param = config_model.addStringParameter2( "offlinedownloader.device.name", "offlinedownloader.device.name", service_name );
			
			service_name = sn_param.getValue();
			
			BooleanParameter slave_param = config_model.addBooleanParameter2( "offlinedownloader.device.slave", "offlinedownloader.device.slave", false );

			slave_mode = slave_param.getValue();
			
			slave_param.addListener(
				new ParameterListener() 
				{
					public void 
					parameterChanged(
						Parameter param )
					{
						slave_mode = ((BooleanParameter)param).getValue();
					}
				});
			
			DownloadManager	download_manager = plugin_interface.getDownloadManager();
			
			download_manager.addListener(
				new DownloadManagerListener()
				{
					public void
					downloadAdded(
						Download	download )
					{
						String client_id = download.getAttribute( ta_category );
						
						if ( client_id != null ){
					
							setPiecePriorities( download );
							
							if ( download.getState() != Download.ST_STOPPED && !slave_mode ){
								
								try{
									log( "Stopping " + download.getName() + " as found to be running on startup" );
															
									download.stop();
									
									deactivateAnnounce( download );

								}catch( Throwable e ){
									
									log( "Failed to stop", e );
								}
							}
							
								// mark all clients as active as at this point we don't know whether they are or not
							
							if ( !active_clients.containsKey( client_id )){
								
								log( "Adding client " + client_id );
								
								clientState state = new clientState( client_id );
								
								state.setLastSeen( SystemTime.getMonotonousTime());
								
								active_clients.put( client_id, state );
							}
						}
						
						download.addPeerListener(
							new DownloadPeerListener()
						{
								public void
								peerManagerAdded(
									final Download		download,
									PeerManager			peer_manager )
								{
									PeerManagerImpl	_pm = (PeerManagerImpl)peer_manager;
									
									final PEPeerManager pm = _pm.getDelegate();
									
									pm.getPiecePicker().addPriorityProvider(
										new PiecePriorityProvider()
										{
											public long[]
											updatePriorities(
												PiecePicker		picker )
											{
												return( updatePiecePriorities( download, picker.getNumberOfPieces()));
											}
										});
								}
								
								public void
								peerManagerRemoved(
									Download		download,
									PeerManager		peer_manager )
								{
									
								}
						});
					}
					
					public void
					downloadRemoved(
						Download	download )
					{
						removeExplicitStart( download );
						
						synchronized( download_priorities_map ){
							
							download_priorities_map.remove( download );
						}
						
						synchronized( xfer_priorities_map ){
							
							xfer_priorities_map.remove( download );
						}
					}
				}, true );
			
			plugin_interface.addListener(
					new PluginListener()
					{
						public void
						initializationComplete()
						{
							DelayedTask dt = 
								plugin_interface.getUtilities().createDelayedTask(new Runnable()
								{
									public void 
									run()
									{
										Thread t = 
											new Thread( "OfflineDownloader::init" )
											{
												public void
												run()
												{
													OfflineDownloaderPlugin.this.start();
												}
											};
										
										t.setPriority( Thread.MIN_PRIORITY );
										
										t.setDaemon( true );
										
										t.start();
									}
								});
							
							dt.queue();
						}
						
						public void
						closedownInitiated()
						{	
							if ( ssdp != null ){
								
								stop();
							}
						}
						
						public void
						closedownComplete()
						{
						}
					});
			
		}catch( Throwable e ){
			
			Debug.out( e );
		}
	}
	
	
	protected void
	start()
	{
		synchronized( startup_lock ){
			
			if ( starting ){
				
				return;
			}
			
			starting	= true;
		}
		
		logger.log( "Server starts: upnp enabled=" + enable_upnp );
				
		Utilities utils = plugin_interface.getUtilities();

		UTTimer timer 	= utils.createTimer( "ODTimer", true );

		timer.addPeriodicEvent(
				ACTIVE_CHECK_PERIOD,
				new UTTimerEventPerformer()
				{
					public void
					perform(
						UTTimerEvent		event )
					{
						checkActivation();
					}
				});
		
		try{			
			if ( enable_upnp ){
				
				UPnPAdapter adapter = 
					new UPnPAdapter()
					{
						public SimpleXMLParserDocument
						parseXML(
							String	data )
						
							throws SimpleXMLParserDocumentException
						{
							return( plugin_interface.getUtilities().getSimpleXMLParserDocumentFactory().create( data ));
						}
						
						public ResourceDownloaderFactory
						getResourceDownloaderFactory()
						{
							return( plugin_interface.getUtilities().getResourceDownloaderFactory());
						}
						
						public UTTimer
						createTimer(
							String	name )
						{
							return( plugin_interface.getUtilities().createTimer( name ));
						}
						
						public void
						createThread(
							String				name,
							final Runnable		runnable )
						{
							plugin_interface.getUtilities().createThread( name, runnable );
						}
						
						public Comparator
						getAlphanumericComparator()
						{
							return( plugin_interface.getUtilities().getFormatters().getAlphanumericComparator( true ));
						}
	
						public void
						log(
							Throwable	e )
						{
							Debug.printStackTrace(e);
						}
						
						public void
						trace(
							String	str )
						{
							// System.out.println( str );
						}
						
						public void
						log(
							String	str )
						{
							// System.out.println( str );
						}
						
						public String
						getTraceDir()
						{
							return( plugin_interface.getPluginDirectoryName());
						}
					};
					
				PluginConfig	config = plugin_interface.getPluginconfig();

				NetworkAdmin	admin = NetworkAdmin.getSingleton();
				
				TrackerWebContext ssdp_web_context;
				
				int	count 		= 0;
				int	ssdp_port	= 0;
				
				while( true ){
					
					count++;
					
					int	port = config.getPluginIntParameter( "ssdp_port", 0 );
										
					ssdp_port = admin.getBindablePort( port );
						
					if ( ssdp_port != port ){
						
						config.setPluginParameter( "ssdp_port", ssdp_port );
					}
					
					try{
						ssdp_web_context = plugin_interface.getTracker().createWebContext(
								ssdp_port,
								Tracker.PR_HTTP );
						
						break;
						
					}catch( Throwable e ){
						
						if ( count >= 8 ){
							
							throw( e );
						}
					}
				}
				
				ssdp = UPnPFactory.getSSDP( 
									adapter, 
									UPnPSSDP.SSDP_GROUP_ADDRESS, 
									UPnPSSDP.SSDP_GROUP_PORT, 
									ssdp_port,
									null );
				
				logger.log( "SSDP port: " + ssdp_port );
								
				
				ssdp_web_context.addPageGenerator(
					new TrackerWebPageGenerator()
					{
						public boolean
						generate(
							TrackerWebPageRequest		request,
							TrackerWebPageResponse		response )
						
							throws IOException
						{
							String	url = request.getURL();
							
							String	header	= request.getHeader();
														
							Map		headers = request.getHeaders();
														
							String	action = (String)headers.get( "soapaction" );					
								
							if ( url.equals( "/VuzeOfflineDownloaderService/Event" )){		
									
								// System.out.println( request.getHeader());
	
								return( false );
								
							}else if ( action == null ){
								
									// unfortunately java lets getSourceAsStream escape from the
									// package hierarchy and onto the file system, at least when
									// loading the class from FS instead of .jar
								
									// first we don't support ..
								
								url = url.trim();
								
								if ( url.indexOf( ".." ) == -1 && !url.endsWith( "/" )){
								
										// second we don't support nested resources
									
									int	 pos = url.lastIndexOf("/");
									
									if ( pos != -1 ){
										
										url = url.substring( pos );
									}
									
									String resource = "/com/aelitis/azureus/plugins/offlinedownloader/resources" + url;
						
									InputStream stream = getClass().getResourceAsStream( resource );
									
									if ( stream != null ){
										
										try{
											if ( url.startsWith( "/RootDevice" )){
												
												byte[]	buffer = new byte[1024];
												
												String	content = "";
												
												while( true ){
													
													int	len = stream.read( buffer );
													
													if ( len <= 0 ){
														
														break;
													}
													
													content += new String( buffer, 0, len, "UTF-8" );
												}
												
												content = content.replaceAll( "%UUID%", UUID_rootdevice );
												content = content.replaceAll( "%SERVICENAME%", service_name );
												
												stream.close();
												
												stream = new ByteArrayInputStream( content.getBytes( "UTF-8" ));
											}
											
											response.useStream( "xml", stream );
										
											return( true );
											
										}finally{
												
											stream.close();
										}
									}
								}
								
								String 	temp = header.replaceAll( "\r", "\\\\r" ).replaceAll( "\n", "\\\\n" );

								logger.log( "HTTP: no match for " + url );
								logger.log( "    header: " + temp );
								logger.log( "    decoded: " + headers );
								
								return( false );
								
							}else{
													
								String	host = (String)headers.get( "host" );
								
								if ( host == null ){
									
									host = "127.0.0.1";
									
								}else{
									
									int	pos = host.indexOf(':');
									
									if ( pos != -1 ){
										
										host	= host.substring(0,pos);
									}
									
									host	= host.trim();
								}
								
								String	client = (String)headers.get( "user-agent" );
								
								if ( client == null ){
									
									client = request.getClientAddress();
									
								}else{
									
									client = client + ":" + request.getClientAddress();
								}
								
								try{									
									String	action_content = "";
									
									byte[]	buffer = new byte[1024];
									
									InputStream	is = request.getInputStream();
									
									while( true ){
										
										int	len = is.read( buffer );
										
										if ( len < 0 ){
											
											break;
										}
										
										action_content += new String( buffer, 0, len, "UTF-8" );
									}
									
									action_content	= action_content.trim();
									
									SimpleXMLParserDocument doc = plugin_interface.getUtilities().getSimpleXMLParserDocumentFactory().create(	action_content );

									String result = processAction( client, url, doc );
										
									if ( result == null ){
										
											// not implemented
										
										response.setReplyStatus( 501 );
										
									}else{
									
										response.setHeader( "SERVER", getServerName());
										
										response.useStream( "xml", new ByteArrayInputStream( result.getBytes( "UTF-8" )));
									}
									
									return( true );
									
								}catch( Throwable e ){
									
									Debug.printStackTrace(e);
									
									logger.log( e );
									
									return( false );
								}
							}
						}
					});
				
				ssdp_listener = 
						new UPnPSSDPListener()
						{
							public void
							receivedResult(
								NetworkInterface	network_interface,
								InetAddress			local_address,
								InetAddress			originator,
								String				USN,
								URL					location,
								String				ST,
								String				AL )
							{
								// System.out.println( "receivedResult: " + location + "/" + ST + "/" + AL );
							}
							
							public void
							receivedNotify(
								NetworkInterface	network_interface,
								InetAddress			local_address,
								InetAddress			originator,
								String				USN,
								URL					location,
								String				NT,
								String				NTS )
							{
								// System.out.println( "receivedNotify: " + USN + "/" + location + "/" + NT + "/" + NTS );					
							}
	
							public String[]
							receivedSearch(
								NetworkInterface	network_interface,
								InetAddress			local_address,
								InetAddress			originator,
								String				ST )
							{
								for (int i=0;i<upnp_entities.length;i++){
									
									if ( ST.equals( upnp_entities[i] )){
										
										return( new String[]{ 
												UUID_rootdevice,
												"RootDevice.xml" });
									}
								}
								
								return( null );
							}
							
							public void
							interfaceChanged(
								NetworkInterface	network_interface )
							{	
							}
						};
				
				ssdp.addListener( ssdp_listener );

				sendAlive();
	
				timer.addPeriodicEvent(
						UPNP_ALIVE_PERIOD,
						new UTTimerEventPerformer()
						{
							public void
							perform(
								UTTimerEvent		event )
							{
								sendAlive();
							}
						});
		
			}else{
			}
		}catch( Throwable e ){
			
		}finally{
			
			startup_sem.releaseForever();
		}
	}
	
	protected String
	processAction(
		String						client,
		String						url,
		SimpleXMLParserDocument		doc )
	{
		SimpleXMLParserDocumentNode	body = doc.getChild( "Body" );

		SimpleXMLParserDocumentNode	call = body.getChildren()[0];
		
		final String	command = call.getName();				
		
		SimpleXMLParserDocumentNode[]	args = call.getChildren();
		
		final Map<String,String>	arg_map = new HashMap<String,String>();
		
		String	arg_str = "";
		
		for (int i=0;i<args.length;i++){
			
			SimpleXMLParserDocumentNode	arg = args[i];
			
			arg_map.put( arg.getName().trim(), arg.getValue().trim());
			
			arg_str += ", " + arg.getName() + " -> " + arg.getValue();
		}
	
		final resultEntries	result = new resultEntries();

		logger.log( "Action (client=" + client + "): " + url + ":" + command + arg_str );
		
		final boolean[] ok = {false};
		
		final AESemaphore sem = new AESemaphore( "exec" );
		
		async_dispatcher.dispatch(
			new AERunnable()
			{
				public void
				runSupport()
				{
					try{
						if ( command.equals( "GetFreeSpace" )){
							
							ok[0] = getFreeSpace( arg_map, result );
							
						}else if ( command.equals( "Activate" )){
							
							ok[0] = activate( arg_map, result );
								
						}else if ( command.equals( "SetDownloads" )){
							
							ok[0] = setDownloads( arg_map, result );
								
						}else if ( command.equals( "AddDownload" )){
							
							ok[0] = addDownload( arg_map, result );
							
						}else if ( command.equals( "AddDownloadChunked" )){
							
							ok[0] = addDownloadChunked( arg_map, result );
							
						}else if ( command.equals( "RemoveDownload" )){
							
							ok[0] = removeDownload( arg_map, result );
							
						}else if ( command.equals( "StartDownload" )){
							
							ok[0] = startDownload( arg_map, result );
							
						}else if ( command.equals( "UpdateDownload" )){
							
							ok[0] = updateDownload( arg_map, result );
						}
					}finally{
						
						sem.release();
					}
				}
			});
		
		sem.reserve();
		
		if ( !ok[0] ){
			
			return( null );
		}
		
		String	response =
			"<?xml version=\"1.0\" encoding=\"utf-8\"?>"+
			"<s:Envelope xmlns:s=\"http://schemas.xmlsoap.org/soap/envelope/\" s:encodingStyle=\"http://schemas.xmlsoap.org/soap/encoding/\">"+
				"<s:Body>";
				
			// todo we should support faults here for errors - see http://www.upnp.org/specs/arch/UPnP-DeviceArchitecture-v1.0.pdf page 50
		
		response += "<u:" + command + "Response xmlns:u=\"urn:schemas-upnp-org:service:VuzeOfflineDownloaderService:1\">";
			
		List<String>	result_args = result.getArgs();
		List<String>	result_vals	= result.getValues();
		
		for (int i=0;i<result_args.size();i++){
			
			String	arg = (String)result_args.get(i);
			String	val	= (String)result_vals.get(i);
			
			response += "<" + arg + ">" + val + "</" + arg + ">";
		}
		
		response += "</u:" + command + "Response>";

		response += "</s:Body>"+
					"</s:Envelope>";
		
		return( response );
	}
	
	
	protected boolean
	getFreeSpace(
		Map<String,String>		args,
		resultEntries			result )
	{
		String	client_id 	= args.get( "NewClientID" );

		if ( client_id == null ){
			
			return( false );
		}
		
		activate( client_id );

		String save_dir = plugin_interface.getPluginconfig().getCoreStringParameter( PluginConfig.CORE_PARAM_STRING_DEFAULT_SAVE_PATH );
		
		String	space;
		
		File f = new File( save_dir );
		
		if ( !f.exists() || !f.canWrite()){
			
			space = "0";
			
		}else{
			
			space = String.valueOf( f.getUsableSpace());
		}
		
		result.set( "NewFreeSpace", space );
		
		return( true );
	}
	
	protected boolean
	activate(
		Map<String,String>		args,
		resultEntries			result )
	{
		String	client_id 	= args.get( "NewClientID" );
		
		if ( client_id == null ){
			
			return( false );
		}
		
		deactivate( client_id );
		
		result.set( "NewStatus", "OK" );
		
		return( true );
	}
	
	protected boolean
	setDownloads(
		Map<String,String>		args,
		resultEntries			result )
	{
		String	client_id 	= args.get( "NewClientID" );
		String	hashes_str	= args.get( "NewTorrentHashList" );
		
		log( "    client_id=" + client_id + ", hashes=" + hashes_str );
		
		if ( client_id == null || hashes_str == null ){
			
			return( false );
		}
		
		activate( client_id );

		String[] hashes = hashes_str.split( "," );
		
		String	result_list = "";
		
		DownloadManager dm = plugin_interface.getDownloadManager();
		
		Set<Download>	matched = new HashSet<Download>();
		
		for ( String hash: hashes ){
			
			hash = hash.trim();
			
			if ( hash.length() == 0 ){
				
				continue;
			}
			
			int	state;
			
			try{
				Download download = dm.getDownload( ByteFormatter.decodeString( hash ));
							
				if ( download == null ){
					
					state = 1;
					
				}else{
					String	category = download.getAttribute( ta_category );
					
					if ( category != null && category.equals( client_id )){
						
						matched.add( download );
					}
					
					state = 0;
				}
			}catch( Throwable e ){
				
				state = 3;
			}
			
			result_list += (result_list.length()==0?"":",") + String.valueOf( state );
		}		
			
			// remove any that are missing
		
		for ( Download download: dm.getDownloads()){
		
			if ( matched.contains( download )){
				
				continue;
			}
			
			String	category = download.getAttribute( ta_category );
			
			if ( category != null && category.equals( client_id )){

				if ( slave_mode ){
					
					download.setAttribute( ta_category, CATEGORY_SERVER_ONLY );
					
				}else{
					
					int	state = download.getState();
					
					try{
						if ( state == Download.ST_ERROR || state == Download.ST_STOPPED ){
							
							log( "Removing " + download.getName() + " as client no longer has it running" );
							
							download.remove( true, true );
							
						}else{
							
							log( "Stopping " + download.getName() + " as client no longer has it running" );
							
							download.stop();
						}
					}catch( Throwable e ){
						
						log( "Failed to decommision download", e );
					}
				}
			}
		}
		
		result.set( "NewSetDownloadsResultList", result_list );
		
		result.set( "NewStatus", "OK" );

		return( true );
	}
	
	protected boolean
	addDownload(
		Map<String,String>		args,
		resultEntries			result )
	{
		String	client_id 	= args.get( "NewClientID" );
		String	hash		= args.get( "NewTorrentHash" );
		String	data		= args.get( "NewTorrentData" );
		
		log( "    client_id=" + client_id + ", hash=" + hash + ", data=" + (data==null?"null":String.valueOf( data.length())));
		
		if ( client_id == null || hash == null || data == null){
			
			return( false );
		}
		
		activate( client_id );

		try{
			TOTorrent t = TOTorrentFactory.deserialiseFromMap( BDecoder.decode( ByteFormatter.decodeString( data )));
		
			Download download = 
				plugin_interface.getDownloadManager().addDownloadStopped( 
						PluginCoreUtils.wrap( t ), null, null );
			
			download.setAttribute( ta_category, client_id );
			
			if ( slave_mode ){
				
				download.restart();
				
			}else{
			
				deactivateAnnounce( download );
			}
			
			result.set( "NewStatus", "OK" );
		
		}catch( Throwable e ){
						
			result.set( "NewStatus", "ERROR_OPERATION_FAILED" );
			
			log( "Failed to add download", e );
		}
		return( true );
	}
		
	private String 	current_chunked_hash 	= "";
	private String	current_chunked_data	= "";
	
	protected boolean
	addDownloadChunked(
		Map<String,String>		args,
		resultEntries			result )
	{
		String	client_id 		= args.get( "NewClientID" );
		String	hash			= args.get( "NewTorrentHash" );
		int		offset			= Integer.parseInt( args.get( "NewChunkOffset" ));
		int		total_length	= Integer.parseInt( args.get( "NewTotalLength" ));
		String	data			= args.get( "NewTorrentData" );
		
		log( "    client_id=" + client_id + ", hash=" + hash + ", offset=" + offset + ", total=" + total_length + ", data=" + (data==null?"null":String.valueOf( data.length())));
		
		if ( client_id == null || hash == null || data == null){
			
			return( false );
		}
		
		activate( client_id );

		try{
			if ( !current_chunked_hash.equals( hash )){
				
				current_chunked_hash 	= hash;
				current_chunked_data	= "";
			}

			if ( offset != current_chunked_data.length()){
				
				throw( new Exception( "Chunked data offset mismatch (expected=" + current_chunked_data.length() + ", supplied=" + offset + ")" ));
			}
			
			current_chunked_data += data;
			
			if ( current_chunked_data.length() == total_length ){
			
				TOTorrent t = TOTorrentFactory.deserialiseFromMap( BDecoder.decode( ByteFormatter.decodeString( current_chunked_data )));
			
				current_chunked_data 	= "";
				current_chunked_hash	= "";
				
				Download download = 
					plugin_interface.getDownloadManager().addDownloadStopped( 
							PluginCoreUtils.wrap( t ), null, null );
				
				download.setAttribute( ta_category, client_id );
				
				if ( slave_mode ){
					
					download.restart();
					
				}else{
					
					deactivateAnnounce( download );
				}
				
			}else if ( current_chunked_data.length() > total_length ){
			
				throw( new Exception( "Chunked data offset mismatch (too large)" ));
			}
			
			result.set( "NewStatus", "OK" );
		
		}catch( Throwable e ){
						
			result.set( "NewStatus", "ERROR_OPERATION_FAILED" );
			
			log( "Failed to add download", e );
		}
		return( true );
	}
	
	protected boolean
	updateDownload(
		Map<String,String>		args,
		resultEntries			result )
	{
		String	client_id 	= args.get( "NewClientID" );
		String	hash		= args.get( "NewTorrentHash" );
		String	required	= args.get( "NewPieceRequiredMap" );
		
		log( "    client_id=" + client_id + ", hash=" + hash + ", required=" + (required==null?"null":String.valueOf( required.length())));
		
		if ( client_id == null || hash == null || required == null){
			
			return( false );
		}
		
		activate( client_id );

		result.set( "NewPieceHaveMap", "" );

		try{
			DownloadManager dm = plugin_interface.getDownloadManager();

			Download download = dm.getDownload( ByteFormatter.decodeString( hash ));
			
			if ( download == null ){
				
				result.set( "NewStatus", "ERROR_UNKNOWN_DOWNLOAD" );
				
			}else{
				
				String	category = download.getAttribute( ta_category );
				
				if ( category != null && category.equals( client_id )){
				
					byte[] required_map = ByteFormatter.decodeString( required );
					
					setPiecePriorities( download, required_map );
					
					byte[] have_map = getHavePieceMap( download );
					
						// if we don't have any interesting pieces then kill explicit start if in operation
					
					int	useful_piece_count = 0;
					
					if ( have_map != null ){
						
						for ( int i=0;i<have_map.length; i++ ){
							
							int x = ( have_map[i] & required_map[i] )&0xff;
							
							if ( x != 0 ){
								
								for (int j=0;j<8;j++){
									
									if ((x&0x01) != 0 ){
										
										useful_piece_count++;
									}
									
									x >>= 1;
								}
							}
						}
					}
					
					log( "    we have " + useful_piece_count + " pieces that the client doesn't" );
					
					if ( useful_piece_count == 0 ){
						
						removeExplicitStart( download );
					}
					
					result.set( "NewPieceHaveMap", have_map==null?"": ByteFormatter.encodeStringFully( have_map ));
				}
				
				result.set( "NewStatus", "OK" );
			}
		}catch( Throwable e ){
			
			log( "Failed to update download", e );
						
			result.set( "NewStatus", "ERROR_OPERATION_FAILED" );
		}
		
		return( true );
	}
	
	protected boolean
	removeDownload(
		Map<String,String>		args,
		resultEntries			result )
	{
		String	client_id 	= args.get( "NewClientID" );
		String	hash		= args.get( "NewTorrentHash" );
		
		log( "    client_id=" + client_id + ", hash=" + hash );
		
		if ( client_id == null || hash == null ){
			
			return( false );
		}
		
		activate( client_id );

		try{
			DownloadManager dm = plugin_interface.getDownloadManager();

			Download download = dm.getDownload( ByteFormatter.decodeString( hash ));
			
			if ( download == null ){
				
				result.set( "NewStatus", "ERROR_UNKNOWN_DOWNLOAD" );
				
			}else{
				
				String	category = download.getAttribute( ta_category );
				
				if ( category != null && category.equals( client_id )){
				
					if ( slave_mode ){
						
						download.setAttribute( ta_category, CATEGORY_SERVER_ONLY );
						
					}else{
						
						int	state = download.getState();
						
						try{
							if ( state == Download.ST_ERROR || state == Download.ST_STOPPED ){
								
								log( "Removing " + download.getName() + " as client requested" );
								
								download.remove( true, true );
								
							}else{
								
								log( "Stopping " + download.getName() + " as client requested removal" );
								
								download.stop();
							}
						}catch( Throwable e ){
							
							log( "Failed to decommision download", e );
						}
					}
				}
				
				result.set( "NewStatus", "OK" );
			}
		}catch( Throwable e ){
			
			log( "Failed to remove download", e );
						
			result.set( "NewStatus", "ERROR_OPERATION_FAILED" );
		}
		
		return( true );
	}
	
	protected boolean
	startDownload(
		Map<String,String>		args,
		resultEntries			result )
	{
		String	client_id 	= args.get( "NewClientID" );
		String	hash		= args.get( "NewTorrentHash" );
		
		log( "    client_id=" + client_id + ", hash=" + hash );
		
		if ( client_id == null || hash == null ){
			
			return( false );
		}
		
		activate( client_id );

		result.set( "NewDataPort", "0" );
		
		try{
			DownloadManager dm = plugin_interface.getDownloadManager();

			Download download = dm.getDownload( ByteFormatter.decodeString( hash ));
			
			if ( download == null ){
				
				result.set( "NewStatus", "ERROR_UNKNOWN_DOWNLOAD" );
				
			}else{
				
				String	category = download.getAttribute( ta_category );
				
				if ( category != null && category.equals( client_id )){
				
					addExplicitStart( download );
				}
				
				result.set( "NewDataPort", String.valueOf( plugin_interface.getPluginconfig().getCoreIntParameter( PluginConfig.CORE_PARAM_INT_INCOMING_TCP_PORT )));
				
				result.set( "NewStatus", "OK" );
			}
		}catch( Throwable e ){
			
			log( "Failed to start download", e );
						
			result.set( "NewStatus", "ERROR_OPERATION_FAILED" );
		}
		
		return( true );
	}
	
	protected void
	checkActivation()
	{
		synchronized( active_clients ){
		
			long	now = SystemTime.getMonotonousTime();
			
			Iterator<clientState> it = active_clients.values().iterator();
			
			while( it.hasNext()){
			
				clientState	state = it.next();
				
				if ( now - state.getLastSeen() >= ACTIVE_TIMEOUT ){
					
					it.remove();
					
					log( "Deactivated " + state.getID());
				}
			}
		}
								
		checkState();
	}
	
	protected void
	activate(
		String		client_id )
	{
		boolean	changed = false;
		
		synchronized( active_clients ){
		
			clientState state = active_clients.get( client_id );
			
			if ( state == null ){
				
				state = new clientState( client_id );
				
				active_clients.put( client_id, state );
				
				changed = true;
			}
			
			state.setLastSeen( SystemTime.getMonotonousTime());
		}
		
		if ( changed ){
			
			log( "Activated " + client_id );
			
			checkState();
		}
	}

	protected void
	deactivate(
		String		client_id )
	{
		boolean	changed = false;
		
		synchronized( active_clients ){
		
			changed = active_clients.remove( client_id ) != null;
		}
		
		if ( changed ){
			
			log( "Deactivated " + client_id );
			
			checkState();
		}
	}
	
	protected void
	checkState()
	{
		async_dispatcher.dispatch(
			new AERunnable()
			{
				public void
				runSupport()
				{
					checkStateSupport();
				}
			});
	}
	
	protected void
	checkStateSupport()
	{
		Set<String>	active;
		
		synchronized( active_clients ){

			active = new HashSet<String>( active_clients.keySet());
		}
		
		Download[]	downloads = plugin_interface.getDownloadManager().getDownloads();
		
		for ( Download download: downloads ){
			
			String client_id = download.getAttribute( ta_category );
			
			if ( client_id != null ){

				int	state = download.getState();

				if ( active.contains( client_id )){
					
					if ( !isExplicitStart( download ) && !slave_mode ){
					
						if ( state != Download.ST_STOPPED ){
							
							try{
								getHavePieceMap( download );
								
								download.stop();
								
								log( "Stopped " + download.getName());
								
							}catch( Throwable e ){
								
								log( "Failed to stop download " + download.getName(), e );
							}
							
							deactivateAnnounce( download );
						}
					}
				}else{
					
					removeExplicitStart( download );
					
					if ( state == Download.ST_STOPPED ){
												
						activateAnnounce( download );
						
						try{
							download.restart();
							
							log( "Started " + download.getName());
							
						}catch( Throwable e ){
							
							log( "Failed to start download " + download.getName(), e );
						}
					}else{
						
						getHavePieceMap( download );
					}
				}
			}
		}
	}
	
	protected byte[]
	getHavePieceMap(
		Download		download )
	{
		Map	piece_details = download.getMapAttribute( ta_piece_details );
		
		if ( piece_details == null ){
			
			piece_details = new HashMap();
		}
		
		byte[]	old_have_pieces = (byte[])piece_details.get( "h" );
		
		byte[]	new_have_pieces;

		try{
				// when seeding it means we have all the required pieces 
						
			if ( download.isComplete()){
			
				new_have_pieces = (byte[])piece_details.get( "r" );
				
			}else{
			
				DiskManager disk_manager = PluginCoreUtils.unwrap( download ).getDiskManager();
			
				if ( disk_manager != null ){
				
					DiskManagerPiece[] pieces = disk_manager.getPieces();
					
					new_have_pieces = new byte[( pieces.length + 7 ) / 8];
					
					int	have_pos		= 0;
					int	current_byte	= 0;
					int	pos 			= 0;
					
					int	hits = 0;
					
					for ( DiskManagerPiece piece: pieces ){
						
						current_byte = current_byte << 1;
						
						if ( piece.isDone()){
							
							current_byte += 1;
							
							hits++;
						}
						
						if (( pos %8 ) == 7 ){
							
							new_have_pieces[have_pos++] = (byte)current_byte;
							
							current_byte = 0;
						}
						pos++;
					}
					
					if (( pos % 8 ) != 0 ){
						
						new_have_pieces[have_pos++] = (byte)(current_byte << (8 - (pos % 8)));
					}
					
					if ( old_have_pieces == null || !Arrays.equals( new_have_pieces, old_have_pieces )){
						
						Map new_piece_details = new HashMap( piece_details );
						
						new_piece_details.put( "h", new_have_pieces );
						
						download.setMapAttribute( ta_piece_details, new_piece_details );
					}
				}else{
					
					new_have_pieces = old_have_pieces;
				}
			}
		}catch( Throwable e ){
			
			Debug.out( "Failed to getHavePieceMap for " + download.getName(), e );
			
			new_have_pieces = old_have_pieces;
		}
		
		return( new_have_pieces );
	}
	
	protected long[]
	updatePiecePriorities(
		Download		download,
		int				num_pieces )
	{
		String client_id = download.getAttribute( ta_category );

		if ( client_id == null || slave_mode ){
			
			return( null );
		}
		
		if ( isExplicitStart( download )){
			
			synchronized( xfer_priorities_map ){
				
				long[] priorities = xfer_priorities_map.get( download );
				
				if ( priorities == null ){
			
					priorities = new long[ num_pieces ];
					
					Arrays.fill( priorities, Integer.MIN_VALUE );
					
					xfer_priorities_map.put( download, priorities );
				}
				
				return( priorities );
			}
		}else{
			
			synchronized( download_priorities_map ){
				
				long[] priorities = download_priorities_map.get( download );
				
				if ( priorities == null ){
					
					priorities = setPiecePriorities( download );
				}
				
				if ( priorities == null ){
					
					priorities = new long[ num_pieces ];
					
					Arrays.fill( priorities, Integer.MIN_VALUE );
					
					download_priorities_map.put( download, priorities );
				}
				
				return( priorities );
			}
		}
	}
	
	protected long[]
	setPiecePriorities(
		Download		download )
	{
		Map	piece_details = download.getMapAttribute( ta_piece_details );
		
		if ( piece_details == null ){
		
			return( null );
		}
		
		byte[]	required_pieces = (byte[])piece_details.get( "r" );
		
		if ( required_pieces == null ){
			
			return( null );
		}
		
		return( setPiecePriorities( download, required_pieces ));
	}
	
	protected long[]
	setPiecePriorities(
		Download		download,
		byte[]			new_required_pieces )
	{
		Map	piece_details = download.getMapAttribute( ta_piece_details );
		
		if ( piece_details == null ){
		
			piece_details = new HashMap(); 
		}
		
		byte[]	old_required_pieces = (byte[])piece_details.get( "r" );
		
		boolean	changed = old_required_pieces == null || !Arrays.equals( old_required_pieces, new_required_pieces );
		
		if ( changed ){
			
			Map new_piece_details = new HashMap( piece_details );
			
			new_piece_details.put( "r", new_required_pieces );
			
			download.setMapAttribute( ta_piece_details, new_piece_details );
		}
		
		synchronized( download_priorities_map ){
			
			long[]	priorities = download_priorities_map.get( download );
			
			if ( priorities == null || changed ){
				
				if ( priorities == null ){
					
					priorities = new long[(int)download.getTorrent().getPieceCount()];
					
					download_priorities_map.put( download, priorities );
				}
				
				int	pos		= 0;
				
				int	current	= 0;
				
				for (int i=0;i<priorities.length;i++){
				
					if ( i % 8 == 0 ){
						
						current = new_required_pieces[pos++]&0xff;
					}
					
					priorities[i] = (current&0x80)==0?Integer.MIN_VALUE:1;
					
					current <<= 1;
				}
			}
			
			return( priorities );
		}
	}
	
	protected void
	deactivateAnnounce(
		Download		download )
	{
		TOTorrent torrent = PluginCoreUtils.unwrap( download.getTorrent());
		
		Map cache = download.getMapAttribute( ta_announce_cache );
		
		if ( cache == null ){
			
			String	announce_url = torrent.getAnnounceURL().toExternalForm();
		
			List<List<String>> announce_groups = TorrentUtils.announceGroupsToList( torrent );
			
			cache = new HashMap();
			
			try{
				cache.put( "a",announce_url.getBytes( "UTF-8" ));
				
				List<List<byte[]>> temp = new ArrayList<List<byte[]>>();
				
				for (List<String> ls: announce_groups ){
					
					List<byte[]> x = new ArrayList<byte[]>();
					
					temp.add( x );
					
					for (String s: ls ){
						
						x.add( s.getBytes( "UTF-8" ));
					}
				}
				
				cache.put( "g", temp );
				
				download.setMapAttribute( ta_announce_cache, cache );
				
			}catch( Throwable e ){
				
				Debug.out( e );
			}
		}
		
		download.requestTrackerAnnounce( true );
		
		torrent.setAnnounceURL( DUMMY_ANNOUNCE_URL );
		
		torrent.getAnnounceURLGroup().setAnnounceURLSets( new TOTorrentAnnounceURLSet[0] );
	}
	
	protected void
	activateAnnounce(
		Download		download )
	{
		TOTorrent torrent = PluginCoreUtils.unwrap( download.getTorrent());
		
		Map cache = download.getMapAttribute( ta_announce_cache );
		
		if ( cache == null ){
			
			log( "Can't activate announce, cache missing" );
			
			return;
		}
		
		try{
			List<List<byte[]>>	temp = (List<List<byte[]>>)cache.get( "g" );
			
			List<List<String>> announce_groups = new ArrayList<List<String>>();
			
			for ( List<byte[]> lb: temp ){
				
				List<String> x = new ArrayList<String>();
				
				announce_groups.add( x );
				
				for ( byte[] b: lb ){
					
					x.add( new String( b, "UTF-8" ));
				}
			}	
			
			TorrentUtils.listToAnnounceGroups(  announce_groups, torrent );
			
			torrent.setAnnounceURL( new URL( new String( (byte[])cache.get( "a" ), "UTF-8" )));
			
		}catch( Throwable e ){
			
			Debug.out( e );
		}
	}
	
	protected void
	addExplicitStart(
		Download	download )
	{
		Download[]	downloads;
		
		synchronized( explicit_starts ){
			
			if ( explicit_starts.add( download )){
				
				log( "Added explicit download " + download.getName());
			}
			
			downloads = explicit_starts.toArray( new Download[ explicit_starts.size()]);
		}
		
		for ( Download d: downloads ){
		
			if ( d.getState() == Download.ST_STOPPED ){
				
				try{
					d.restart();
					
					log( "Explicitly started " + d.getName());
					
				}catch( Throwable e ){
					
					log( "Explicit start failed for " + d.getName(), e );
				}
			}
		}
	}
	
	protected void
	removeExplicitStart(
		Download	download )
	{
		boolean	stop_it = false;
		
		synchronized( explicit_starts ){
			
			if ( explicit_starts.remove( download )){
				
				stop_it	= true;
				
				log( "Removed explicit download " + download.getName());
			}
		}
		
		if ( stop_it && !slave_mode ){
			
			if ( download.getState() != Download.ST_STOPPED ){
				
				try{
					download.stop();
					
					log( "Explicitly stopped " + download.getName());
					
				}catch( Throwable e ){
					
					log( "Explicit stop failed for " + download.getName(), e );
				}
			}
		}
	}
	
	protected boolean
	isExplicitStart(
		Download	download )
	{
		synchronized( explicit_starts ){

			return( explicit_starts.contains( download ));
		}
	}
	
	protected void
	stop()
	{
		sendDead();
	}
	
	protected void
	sendAlive()
	{
		synchronized( this ){
			
			if ( alive ){
		
				sendNotify( "ssdp:alive" );
			}
		}
	}
	
	protected void
	sendDead()
	{
		synchronized( this ){

			alive	= false;
			
			sendNotify( "ssdp:byebye" );
		}
	}
	
	protected void
	sendNotify(
		String	status )
	{
		if ( ssdp != null ){
			
			for (int i=0;i<upnp_entities.length;i++){
				
				ssdp.notify( upnp_entities[i], status, upnp_entities[i]==UUID_rootdevice?null:UUID_rootdevice, "RootDevice.xml" );
			}
		}
	}
	
	protected String
	getServerName()
	{
		return( System.getProperty( "os.name" ) + "/" + System.getProperty("os.version") + " UPnP/1.0 " +
				Constants.AZUREUS_NAME + "/" + Constants.AZUREUS_VERSION );
	}
	
	protected String
	escapeXML(
		String	str )
	{
		if ( str == null ){
			
			return( "" );
			
		}
		str = str.replaceAll( "&", "&amp;" );
		str = str.replaceAll( ">", "&gt;" );
		str = str.replaceAll( "<", "&lt;" );
		// str = str.replaceAll( "\"", "&quot;" );	breaks Philips streamium...
		
		return( str );
	}
	
	protected void
	log(
		String		str )
	{
		logger.log( str );
	}
	
	protected void
	log(
		String		str,
		Throwable	e )
	{
		logger.log( str, e );
	}
	
		// IPC start
	
	public String
	getUSN()
	{
		return( UUID_rootdevice );
	}
	
		// IPC end
	
	protected class
	resultEntries
	{
		private Map<String,String>	results = new LinkedHashMap<String, String>();
				
		public void
		set(
			String		name,
			String		value )
		{
			results.put( name, escapeXML( value ));
		}
		
		public List<String>
		getArgs()
		{
			return( new ArrayList<String>( results.keySet()));
		}
		
		public List<String>
		getValues()
		{
			return( new ArrayList<String>( results.values()));
		}
	}
	
	protected class
	clientState
	{
		private String		id;
		private long		last_seen;
		
		protected
		clientState(
			String		_id )
		{
			id	= _id;
		}
		
		protected String
		getID()
		{
			return( id );
		}
		
		protected long
		getLastSeen()
		{
			return( last_seen );
		}
		
		protected void
		setLastSeen(
			long	l )
		{
			last_seen = l;
		}
	}
}
